import logging
from sys import stdout, stderr

import colorlog
from colorlog import ColoredFormatter


def init_logger(dunder_name, testing_mode) -> logging.Logger:
    log_format = (
        '%(asctime)s - '
        '%(name)s - '
        '%(funcName)s - '
        '%(levelname)s - '
        '%(message)s'
    )
    bold_seq = '\033[1m'
    colorlog_format = (
        f'{bold_seq} '
        '%(log_color)s '
        f'{log_format}'
    )
    colorlog.basicConfig(format=colorlog_format)
    logger = logging.getLogger(dunder_name)

    if testing_mode:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # Output full log
    fh = logging.StreamHandler(stdout)
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter(log_format)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    # Output error log
    fh = logging.StreamHandler(stderr)
    fh.setLevel(logging.ERROR)
    formatter = logging.Formatter(log_format)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    return logger



import base64
from datetime import datetime
from typing import Callable, List, Dict

from .client import WebSocketClient
from .log import init_logger


class EssConnect:
    ws: WebSocketClient
    callbacks: Dict[str, Callable]

    def __init__(self, settings, status_function):
        self.setMonitoringResponse(status_function)
        creds = settings["username"] + ":" + settings["password"]
        creds = base64.b64encode(creds.encode("utf-8")).decode("utf-8")
        version = "1.0.8"
        app_name = settings["applicationName"]
        options = {
            "headers": {
                "User-Agent": f"ESS-Python-Agent v{version} - {app_name}",
                "Authorization": "Basic " + creds
            },
            "debug": settings["debug"]
        }
        self.callbacks = {}
        self.logger = init_logger(__name__, testing_mode=True)

        self.ws = WebSocketClient(settings["server"], options)

        self.ws.register_callback('close', self.on_close)
        self.ws.register_callback('message', self.on_message)
        self.ws.register_callback('error', self.on_error)
        self.ws.register_callback('timeout', self.on_timeout)

    async def start_client(self):
        await self.ws.start()

    async def on_message(self, _message):
        message_type = _message['type']
        self.logger.debug(_message)
        if message_type == "ping":
            self.logger.debug('responding to connection-ping')
            await self.send('pong', {})
        elif message_type == "authorized":
            self.logger.debug(
                ('connected and authorized for', _message["namespace"]))
            self.callbacks['authorized'](_message["namespace"])
        elif message_type == "system":
            await self.handle_system_message(_message)
        elif message_type == "message":
            self.current_message = _message
            await self.callbacks['message']({
                "content": _message["content"],
                "parsed":  self.parse_message(_message["content"]["data"]),
                "resolve": self.handle_output
            })
        else:
            self.logger.warning(f"Unknown MessageType {message_type}")

    def on_close(self, error):
        self.logger.debug("close has been called")
        self.callbacks['disconnected'](error)

    def on_error(self, error):
        self.logger.error(error)
        self.callbacks['error'](error)

    def on_timeout(self, error):
        self.logger.error('ESS-Timeout', error)

    async def handle_output(self, output_data):
        await self.send('response', {
            "toMessage":  self.current_message,
            "outputData": output_data
        })

    async def send(self, message_type, content):
        try:
            await self.ws.send({
                "type":      message_type,
                "content":   content,
                "timestamp": str(datetime.now())
            })
        except Exception as e:
            self.logger.error(("Error", e))

    async def handle_system_message(self, message):
        if message["content"]["action"] == "ping":
            self.logger.debug("responding to system-ping")
            try:
                await self.send('system', {
                    "action":      'pong',
                    "client_data": self._start_function() if callable(self._start_function) else {"status": 'ok'},
                    "server_data": message["content"]["id"]
                })
            except Exception as e:
                self.logger.error(("Error", e))

    def setMonitoringResponse(self, start_function):
        self._start_function = start_function

    def parse_message(self, message: List[str]):
        return {part['reference']: part['values'] for part in message}

    def register_callback(self, event, callback: Callable) -> None:
        if event in self.callbacks:
            self.logger.debug(f"Replacing callback function for {event} event")
        else:
            self.logger.debug(f"Setting callback function for {event} event")
        self.callbacks[event] = callback



import asyncio
import json
from errno import ECONNREFUSED
from sys import maxsize
from typing import Callable

import websockets
from websockets.http import Headers

from .log import init_logger


class WebSocketClient():

    def __init__(self, url, options):
        self.number = 0
        self.url = url
        self.autoReconnectInterval = 5 
        self.options = options
        self._debug = options['debug'] | True

        self.callbacks = {}
        self.logger = init_logger(__name__, testing_mode=True)
        self.connectme()

    async def message(self, msg):
        self.number += 1
        if self.number >= maxsize:
            self.logger.debug('Resetting message number')
            self.number = 1
        await self.on_message(msg)

    async def error(self, error):
        if error == ECONNREFUSED:
            await self.reconnect()
        else:
            await self.on_error(error)

    async def on_message(self, message):
        self.logger.debug(message)
        message = json.loads(message)
        message_type = 'incoming-message' if message["type"] == 'message' else 'system-message'
        self.logger.debug(message_type, self.number, message)
        await self.callbacks['message'](message)

    async def on_error(self, error):
        self.logger.error("error", error)
        await self.callbacks['error'](error)

    async def on_close(self, error):
        self.callbacks['close'](error)
        await self.reconnect()

    def connectme(self):
        headers = Headers(self.options["headers"])
        self.connection = websockets.connect(self.url, extra_headers=headers)

    async def send(self, data, callback=None):
        data = json.dumps(data)
        try:
            await self.websocket.send(data)
        except Exception as e:
            self.logger.error(e)
            self.callbacks['error'](e)

    async def reconnect(self):
        self.logger.warning(f'Connection failed retry in {self.autoReconnectInterval} s')
        if not self.connection.ws_client.closed:
            try:
                await self.websocket.close()
            except websockets.exceptions.ConnectionClosedError as error:
                self.logger.debug('Connection closed unexpectedly')
                self.on_close(error)
            except websockets.exceptions.ConnectionClosedOK as error:
                self.logger.debug('Connection closed by remote')
                await self.on_close(error)

        await asyncio.sleep(self.autoReconnectInterval)
        self.connectme()
        self.logger.debug("Reconnecting...")
        await self.start()

    def register_callback(self, event, callback: Callable) -> None:
        if event in self.callbacks:
            self.logger.debug(f"Replacing callback function for {event} event")
        else:
            self.logger.debug(f"Setting callback function for {event} event")
        self.callbacks[event] = callback

    async def start(self):

        async with self.connection as websocket:
            self.websocket = websocket
            while True:
                try:
                    message = await self.websocket.recv()

                    await self.message(message)

                except websockets.exceptions.ConnectionClosedError as error:
                    self.logger.debug(f'Connection closed unexpectedly {error}')
                    self.on_close(error)
                except websockets.exceptions.ConnectionClosedOK as error:
                    self.logger.debug(f'Connection closed by remote {error}')
                    await self.on_close(error)



import pathlib

from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="grexxconnect-ess",
    version="0.1.3",
    description="Grexx External System Service Helper",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Grexx",
    author_email="servicedesk@grexx.net",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=("tests",)),
    install_requires=["websockets", "colorlog"],
)
